#include <stdlib.h>
#include <string.h>
#include <clib/dos_protos.h>
#include <clib/exec_protos.h>
#include <clib/utility_protos.h>
#include <pragmas/dos_pragmas.h>
#include <pragmas/utility_pragmas.h>
#include <pragmas/exec_sysbase_pragmas.h>
#include <exec/memory.h>
#include <dos/dos.h>

#define PoolSize 100000
#define TEMPLATE "PATH/A,RENAME=RN,RECOMMENT=RC,QUIET/S"

struct {
		char *path;
		char *rename;
		char *recomment;
		long quiet;
} arg = {0,0,0,0};

struct fibbuf {
		struct fibbuf *nextobject;
		struct fibbuf *nextdir;
		char *name;
		char *note;
		unsigned long size;
		unsigned char namelen;
		long objectype;
};

char *msg[] = {
		"Directory \"%s\" is empty.\n",
		"*** USER BREAK\n",
		"Out of memory\n",
		"Error in PATH\n",
		"Error in arguments\n",
		"Directory scanning error\n",
		"File access error\n",
		"File access error\n",
		"Invalid charset table\n",
		"v37+ Operating System required\n"
};

enum {
		EMPTY,
		ERR_BREAK,
		ERR_OUTOFMEM,
		ERR_INFROMPATH,
		ERR_INARGS,
		ERR_DIRSCAN,
		ERR_OPEN,
		ERR_READ,
		ERR_TABLE,
		ERR_OSVER
};

APTR ScanTree(BPTR);
void print(char *, int);
unsigned long testbreak(void);
void WalkTree(struct fibbuf *, BPTR);

BPTR outhandle;
struct FileInfoBlock fib;
struct Library *UtilityBase;
struct Library *DOSBase;
struct Library *SysBase;
unsigned long memsize, ec = 0, rc = 0;
void *pool;
struct fibbuf *tree;
struct FileHandle *fhp;
__far unsigned char nametable[256];
__far unsigned char notetable[256];

__saveds main()
{
	SysBase = *((struct Library **)4L);
	if(DOSBase = OpenLibrary("dos.library",0))
	{
		if(UtilityBase=OpenLibrary("utility.library",37))
		{
			struct RDArgs *rda;
			if(rda = ReadArgs(TEMPLATE,(long *)&arg,NULL))
			{
				BPTR lockin;
				if(arg.rename)
				{
					if(lockin = Open(arg.rename,MODE_OLDFILE))
					{
						if(256 != Read(lockin,nametable,256))
						{
							ec = ERR_TABLE;
						}
						else
						{
							unsigned char *cch=nametable;
							int cnt=256;
							while(--cnt)
							{
								if(':'==*cch) *cch='_';
								if('/'==*cch) *cch='_';
								if(0==*cch) *cch='_';
								cch++;
							}
						}
						Close(lockin);
					}
					else
					{
						ec = ERR_OPEN;
					}
				}
				if(arg.recomment)
				{
					if(lockin = Open(strcmp(arg.recomment,"*")?arg.recomment:arg.rename,MODE_OLDFILE))
					{
						if(256 != Read(lockin,notetable,256))
						{
							ec = ERR_TABLE;
						}
						Close(lockin);
					}
					else
					{
						ec = ERR_OPEN;
					}
				}
				if(!ec)
				{
					if(lockin = Lock(arg.path,ACCESS_READ))
					{
						if(pool = CreatePool(MEMF_CLEAR,PoolSize,PoolSize))
						{
							if(tree=ScanTree(lockin))
							{
								WalkTree(tree,lockin);
							}
							else
							{
								if(ec)
								{
									print(msg[ec],10);
								}
								else
								{
									Printf(msg[EMPTY],arg.path);
								}
							}
							DeletePool(pool);
						}
						else
						{
							print(msg[ERR_OUTOFMEM],10);
						}
						UnLock(lockin);
					}
					else
					{
						print(msg[ERR_INFROMPATH],10);
					}
				}
				else
				{
					print(msg[ec],10);
				}
				FreeArgs(rda);
			}
			else
			{
				print(msg[ERR_INARGS],20);
			}
			CloseLibrary(UtilityBase);
		}
		else
		{
			print(msg[ERR_OSVER],20);
		}
	}
	return(rc);
}


unsigned long testbreak(void)
{
	if(SetSignal(0L,SIGBREAKF_CTRL_C) & SIGBREAKF_CTRL_C)
		ec = ERR_BREAK;
	return(ec);
}

void print(char *msg,int err)
{
	if(!arg.quiet)
		Printf(msg);
	if(err)
		rc = err;
}

void WalkTree(struct fibbuf *inbuf, BPTR inlock)
{
	struct fibbuf *buf;
	BPTR oldlock = CurrentDir(inlock);
	for(buf=inbuf;buf;buf=buf->nextobject)
	{
		if((buf->objectype>0) && (buf->nextdir))
		{
			BPTR lockdir;
			if(lockdir = Lock(buf->name,ACCESS_READ))
			{
				WalkTree(buf->nextdir,lockdir);
				UnLock(lockdir);
				if(ec) break;
			}
			else
			{
				Printf("Can't lock %s\n",buf->name);
			}
		}
		if(arg.recomment)
		{
			unsigned char combuf[82];
			unsigned char *chout = combuf;
			unsigned char *chin = buf->note;
			unsigned char ch;
			strcpy(combuf,buf->note);
			while(ch=*chin++) *chout++=notetable[ch];
			if(strcmp(combuf,buf->note)) SetComment(buf->name,combuf);
		}
		if(arg.rename)
		{
			unsigned char nambuf[108];
			unsigned char *chout = nambuf;
			unsigned char *chin = buf->name;
			unsigned char ch;
			strcpy(nambuf,buf->name);
			while(ch=*chin++) *chout++=nametable[ch];
			if(strcmp(nambuf,buf->name)) Rename(buf->name,nambuf);
		}
	}
	CurrentDir(oldlock);
}

APTR ScanTree(BPTR lock)
{
	struct fibbuf *firstbuf = NULL, *lastbuf = NULL, *buf;
	BPTR lockdir, oldlock;
	unsigned long names, notes;
	if(Examine(lock,&fib))
	{
		while(ExNext(lock,&fib))
		{
			if(testbreak())
				return(NULL);
			names = strlen(fib.fib_FileName)+1;
			notes = strlen(fib.fib_Comment)+1;
			if(buf = AllocPooled(pool,names+notes+sizeof(struct fibbuf)))
			{
				if(!firstbuf)
					firstbuf = buf;
				if(lastbuf)
					lastbuf->nextobject = buf;
				lastbuf = buf;
				buf->size = fib.fib_Size;
				buf->objectype = fib.fib_DirEntryType;
				buf->namelen = names-1;
				buf->note = (buf->name = (char *)(buf+1)) + names;
				CopyMem(fib.fib_FileName,buf->name,names);
				CopyMem(fib.fib_Comment,buf->note,notes);
			}
			else
			{
				ec = ERR_OUTOFMEM;
				return(NULL);
			}
		}
	}
	if(ERROR_NO_MORE_ENTRIES!=IoErr())
	{
		ec = ERR_DIRSCAN;
		return(NULL);
	}
	if(buf = firstbuf)
	{
		oldlock = CurrentDir(lock);
		//nesting++;
		do
		{
			if(buf->objectype>0)
			{
				if(lockdir = Lock(buf->name,ACCESS_READ))
				{
					buf->nextdir = ScanTree(lockdir);
					UnLock(lockdir);
					if(ec)
						break;
				}
				else
				{
					ec = ERR_DIRSCAN;
					break;
				}
			}
		}
		while(buf = buf->nextobject);
		CurrentDir(oldlock);
		if(ec)
			return(NULL);
		//nesting--;
	}
	return(firstbuf);
}
